# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)
import hashlib
import re
from contextlib import contextmanager
from pathlib import Path
from typing import Union

import pytest
from dateutil.parser import parse as parse_date

from scitacean import PID, Client, Dataset, DatasetType, File, IntegrityError
from scitacean.filesystem import RemotePath
from scitacean.logging import logger_name
from scitacean.model import DataFile, OrigDatablock, RawDataset
from scitacean.testing.transfer import FakeFileTransfer


def _checksum(data: bytes) -> str:
    checksum = hashlib.new("md5")
    checksum.update(data)
    return checksum.hexdigest()


@pytest.fixture
def data_files():
    contents = {
        "file1.dat": b"contents-of-file1",
        "log/what-happened.log": b"ERROR Flux is off the scale",
        "thaum.dat": b"0 4 2 59 330 2314552",
    }
    files = [
        DataFile(path=name, size=len(content), chk=_checksum(content))
        for name, content in contents.items()
    ]
    return files, contents


@pytest.fixture
def dataset_and_files(data_files):
    model = RawDataset(
        contactEmail="p.stibbons@uu.am",
        creationTime=parse_date("1995-08-06T14:14:14"),
        owner="pstibbons",
        ownerGroup="faculty",
        pid=PID(prefix="UU.000", pid="5125.ab.663.8c9f"),
        principalInvestigator="m.ridcully@uu.am",
        sourceFolder=RemotePath("/src/stibbons/774"),
        type=DatasetType.RAW,
    )
    block = OrigDatablock(
        ownerGroup="faculty",
        size=sum(f.size for f in data_files[0]),
        datasetId=PID(prefix="UU.000", pid="5125.ab.663.8c9f"),
        id=PID(prefix="UU.000", pid="0941.66.abff.41de"),
        dataFileList=data_files[0],
    )
    dset = Dataset.from_models(dataset_model=model, orig_datablock_models=[block])
    return dset, {
        dset.source_folder / name: content for name, content in data_files[1].items()
    }


def load(name: Union[str, Path]):
    with open(name, "rb") as f:
        return f.read()


def test_download_files_creates_local_files_select_all(fs, dataset_and_files):
    dataset, contents = dataset_and_files
    client = Client.without_login(
        url="/", file_transfer=FakeFileTransfer(fs=fs, files=contents)
    )
    client.download_files(dataset, target="./download", select=True)
    assert load("download/file1.dat") == contents["/src/stibbons/774/file1.dat"]
    assert (
        load("download/log/what-happened.log")
        == contents["/src/stibbons/774/log/what-happened.log"]
    )
    assert load("download/thaum.dat") == contents["/src/stibbons/774/thaum.dat"]


def test_download_files_creates_local_files_select_none(fs, dataset_and_files):
    dataset, contents = dataset_and_files
    client = Client.without_login(
        url="/", file_transfer=FakeFileTransfer(fs=fs, files=contents)
    )
    client.download_files(dataset, target="./download", select=False)
    assert not Path("download/file1.dat").exists()
    assert not Path("download/log/what-happened.log").exists()
    assert not Path("download/thaum.dat").exists()


def test_download_files_creates_local_files_select_one_by_string(fs, dataset_and_files):
    dataset, contents = dataset_and_files
    client = Client.without_login(
        url="/", file_transfer=FakeFileTransfer(fs=fs, files=contents)
    )
    client.download_files(dataset, target="./download", select="thaum.dat")
    assert load("download/thaum.dat") == contents["/src/stibbons/774/thaum.dat"]
    assert not Path("download/file1.dat").exists()
    assert not Path("download/log/what-happened.log").exists()


def test_download_files_creates_local_files_select_two_by_string(fs, dataset_and_files):
    dataset, contents = dataset_and_files
    client = Client.without_login(
        url="/", file_transfer=FakeFileTransfer(fs=fs, files=contents)
    )
    client.download_files(
        dataset, target="./download", select=["thaum.dat", "log/what-happened.log"]
    )
    assert not Path("download/file1.dat").exists()
    assert (
        load("download/log/what-happened.log")
        == contents["/src/stibbons/774/log/what-happened.log"]
    )
    assert load("download/thaum.dat") == contents["/src/stibbons/774/thaum.dat"]


def test_download_files_creates_local_files_select_one_by_regex_name_only(
    fs, dataset_and_files
):
    dataset, contents = dataset_and_files
    client = Client.without_login(
        url="/", file_transfer=FakeFileTransfer(fs=fs, files=contents)
    )
    client.download_files(dataset, target="./download", select=re.compile("thaum"))
    assert not Path("download/file1.dat").exists()
    assert not Path("download/log/what-happened.log").exists()
    assert load("download/thaum.dat") == contents["/src/stibbons/774/thaum.dat"]


def test_download_files_creates_local_files_select_two_by_regex_suffix(
    fs, dataset_and_files
):
    dataset, contents = dataset_and_files
    client = Client.without_login(
        url="/", file_transfer=FakeFileTransfer(fs=fs, files=contents)
    )
    client.download_files(dataset, target="./download", select=re.compile(r"\.dat"))
    assert load("download/file1.dat") == contents["/src/stibbons/774/file1.dat"]
    assert not Path("download/log/what-happened.log").exists()
    assert load("download/thaum.dat") == contents["/src/stibbons/774/thaum.dat"]


def test_download_files_creates_local_files_select_one_by_regex_full_path(
    fs, dataset_and_files
):
    dataset, contents = dataset_and_files
    client = Client.without_login(
        url="/", file_transfer=FakeFileTransfer(fs=fs, files=contents)
    )
    client.download_files(
        dataset, target="./download", select=re.compile(r"^file1\.dat$")
    )
    assert load("download/file1.dat") == contents["/src/stibbons/774/file1.dat"]
    assert not Path("download/log/what-happened.log").exists()
    assert not Path("download/thaum.dat").exists()


def test_download_files_creates_local_files_select_one_by_predicate(
    fs, dataset_and_files
):
    dataset, contents = dataset_and_files
    client = Client.without_login(
        url="/", file_transfer=FakeFileTransfer(fs=fs, files=contents)
    )
    client.download_files(
        dataset, target="./download", select=lambda f: f.remote_path == "file1.dat"
    )
    assert load("download/file1.dat") == contents["/src/stibbons/774/file1.dat"]
    assert not Path("download/log/what-happened.log").exists()
    assert not Path("download/thaum.dat").exists()


def test_download_files_returns_updated_dataset(fs, dataset_and_files):
    dataset, contents = dataset_and_files
    client = Client.without_login(
        url="/", file_transfer=FakeFileTransfer(fs=fs, files=contents)
    )
    finalized = client.download_files(dataset, target="./download")

    for f in finalized.files:
        assert f.is_on_local
    for f in dataset.files:
        assert not f.is_on_local  # original is unchanged


def test_download_files_ignores_checksum_if_alg_is_none(fs, dataset_and_files):
    dataset, contents = dataset_and_files

    content = b"random-stuff"
    bad_checksum = "incorrect-checksum"
    model = DataFile(
        path="file.txt",
        size=len(content),
        time=parse_date("2022-06-22T15:42:53+00:00"),
        chk=bad_checksum,
    )
    dataset.add_orig_datablock(checksum_algorithm=None)
    dataset.add_files(File.from_scicat(model))

    client = Client.without_login(
        url="/",
        file_transfer=FakeFileTransfer(
            fs=fs, files={dataset.source_folder / "file.txt": content}
        ),
    )
    # Does not raise
    client.download_files(dataset, target="./download", select="file.txt")


def test_download_files_detects_bad_checksum(fs, dataset_and_files):
    dataset, contents = dataset_and_files

    content = b"random-stuff"
    bad_checksum = "incorrect-checksum"
    model = DataFile(
        path="file.txt",
        size=len(content),
        time=parse_date("2022-06-22T15:42:53+00:00"),
        chk=bad_checksum,
    )
    dataset.add_orig_datablock(checksum_algorithm="md5")
    dataset.add_files(File.from_scicat(model))

    client = Client.without_login(
        url="/",
        file_transfer=FakeFileTransfer(
            fs=fs, files={dataset.source_folder / "file.txt": content}
        ),
    )
    with pytest.raises(IntegrityError):
        client.download_files(dataset, target="./download", select="file.txt")


def test_download_files_detects_bad_size(fs, dataset_and_files, caplog):
    dataset, contents = dataset_and_files

    content = b"random-stuff"
    bad_checksum = hashlib.md5(content).hexdigest()
    model = DataFile(
        path="file.txt",
        size=89412,  # Too large
        time=parse_date("2022-06-22T15:42:53+00:00"),
        chk=bad_checksum,
    )
    dataset.add_orig_datablock(checksum_algorithm="md5")
    dataset.add_files(File.from_scicat(model))

    client = Client.without_login(
        url="/",
        file_transfer=FakeFileTransfer(
            fs=fs, files={dataset.source_folder / "file.txt": content}
        ),
    )
    with caplog.at_level("INFO", logger=logger_name()):
        client.download_files(dataset, target="./download", select="file.txt")
    assert "does not match size reported in dataset" in caplog.text
    assert "89412" in caplog.text


@pytest.mark.skip("Checksum algorithm not yet supported by datablocks")
def test_download_does_not_download_up_to_date_file(fs, dataset_and_files):
    # Ensure the file exists locally
    dataset, contents = dataset_and_files
    client = Client.without_login(
        url="/", file_transfer=FakeFileTransfer(fs=fs, files=contents)
    )
    client.download_files(dataset, target="./download", select=True)

    # Downloading the same files again should not call the downloader.
    class RaisingDownloader(FakeFileTransfer):
        source_dir = "/"

        @contextmanager
        def connect_for_download(self):
            raise RuntimeError("Download disabled")

    client = Client.without_login(
        url="/",
        file_transfer=RaisingDownloader(fs=fs),
    )
    # Does not raise
    client.download_files(dataset, target="./download", select=True)


def test_download_does_not_download_up_to_date_file_manual_checksum(
    fs, dataset_and_files
):
    # Ensure the file exists locally
    dataset, contents = dataset_and_files
    client = Client.without_login(
        url="/", file_transfer=FakeFileTransfer(fs=fs, files=contents)
    )
    client.download_files(dataset, target="./download", select=True)

    # Downloading the same files again should not call the downloader.
    class RaisingDownloader(FakeFileTransfer):
        source_dir = "/"

        @contextmanager
        def connect_for_download(self):
            raise RuntimeError("Download disabled")

    client = Client.without_login(
        url="/",
        file_transfer=RaisingDownloader(fs=fs),
    )
    # Does not raise
    client.download_files(
        dataset, target="./download", select=True, checksum_algorithm="md5"
    )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)
"""Handler for packaged assets."""

import importlib.resources
from functools import lru_cache
from string import Template


def _read_text(filename: str, group: str) -> str:
    if hasattr(importlib.resources, "files"):
        # Use new API added in Python 3.9
        return (
            importlib.resources.files(f"scitacean._html_repr.{group}")
            .joinpath(filename)
            .read_text()
        )
    # Old API, deprecated as of Python 3.11
    # When this is removed, also remove the __init__.py files in the resource folders.
    return importlib.resources.read_text(f"scitacean._html_repr.{group}", filename)


def _preprocess_style(css: str) -> str:
    import re

    # line breaks are not needed
    css = css.replace("\n", "")
    # remove comments
    css = re.sub(r"/\*(\*(?!/)|[^*])*\*/", "", css)
    # remove space around special characters
    css = re.sub(r"\s*([;{}:,])\s*", r"\1", css)
    return css


@lru_cache(maxsize=1)
def dataset_repr_template() -> Template:
    return Template(_read_text("dataset_repr.html.template", "templates"))


@lru_cache(maxsize=1)
def dataset_field_repr_template() -> Template:
    return Template(_read_text("dataset_field_repr.html.template", "templates"))


@lru_cache(maxsize=1)
def files_repr_template() -> Template:
    return Template(_read_text("files_repr.html.template", "templates"))


@lru_cache(maxsize=1)
def metadata_template() -> Template:
    return Template(_read_text("metadata_repr.html.template", "templates"))


@lru_cache(maxsize=1)
def dataset_style() -> str:
    sheet = _preprocess_style(_read_text("dataset.css", "styles"))
    return f"<style>{sheet}</style>"


@lru_cache()
def image(name: str) -> str:
    return _read_text(name, "images")



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)

import hypothesis
import pytest

from .common.backend import scicat_access, scicat_backend  # noqa: F401
from .common.ssh_server import ssh_access, ssh_fileserver  # noqa: F401

# The datasets strategy requires a large amount of memory and time.
# This is not good but hard to avoid.
# So simply disable health checks and accept that tests are slow.
hypothesis.settings.register_profile(
    "scitacean",
    suppress_health_check=[
        hypothesis.HealthCheck.data_too_large,
        hypothesis.HealthCheck.too_slow,
    ],
)


def pytest_addoption(parser: pytest.Parser):
    parser.addoption(
        "--backend-tests",
        action="store_true",
        default=False,
        help="Select whether to run tests against a real SciCat backend",
    )
    parser.addoption(
        "--ssh-tests",
        action="store_true",
        default=False,
        help="Select whether to run tests with an SSH fileserver",
    )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)
"""Load model specification files."""

import dataclasses
from pathlib import Path
from typing import Any, Dict, List, Optional

import yaml

SPEC_DIR = Path(__file__).resolve().parent

DEFAULTS = {
    "default": None,
    "default_factory": None,
    "manual": False,
    "model_name": None,  # equals field name
    "read_only": False,
    "required": False,
    "validation": None,
}


@dataclasses.dataclass
class Field:
    name: str
    default: Optional[Any]
    default_factory: Optional[str]
    description: str
    extra: Optional[dict]
    manual: bool
    model_name: str
    read_only: bool
    required: bool
    type: str
    validation: Optional[str]


@dataclasses.dataclass
class Spec:
    name: str
    inherits: str
    fields: List[Field]


def _load_raw(path: Path):
    with path.open("r") as f:
        return yaml.safe_load(f)


def _load_raw_specs() -> dict:
    return {
        spec["name"]: spec
        for spec in (
            _load_raw(path) for path in SPEC_DIR.iterdir() if path.suffix == ".yml"
        )
    }


def _get_defaults(spec: dict) -> dict:
    return {**DEFAULTS, **spec.get("default_values", {})}


def _apply_defaults_field(field: dict, defaults: dict) -> dict:
    new_field = dict(field)
    for key, default_value in defaults.items():
        if key not in new_field:
            if key == "model_name":
                new_field[key] = new_field["name"]
            else:
                new_field[key] = default_value
    return new_field


def _apply_defaults(spec: dict) -> dict:
    defaults = _get_defaults(spec)
    new_spec = dict(spec)
    new_spec["fields"] = [
        _apply_defaults_field(field, defaults) for field in spec["fields"]
    ]
    return new_spec


def _validate(spec: dict):
    for s in spec.values():
        for field in s["fields"]:
            if field.get("default") and field.get("default_factory"):
                raise ValueError("Cannot use both default and default_factory")


def _to_field_dataclass(field: dict) -> Field:
    field = dict(field)
    args = {
        f.name: field.pop(f.name)
        for f in dataclasses.fields(Field)
        if f.name != "extra"
    }
    extra = field
    return Field(**args, extra=extra)


def _to_spec_dataclasses(specs: dict) -> Dict[str, Spec]:
    return {
        name: Spec(
            name=name,
            inherits=spec.get("inherits", "BaseModel"),
            fields=[_to_field_dataclass(f) for f in spec["fields"]],
        )
        for name, spec in specs.items()
    }


def load_specs() -> Dict[str, Spec]:
    raw_specs = _load_raw_specs()
    defaults_applied = {name: _apply_defaults(spec) for name, spec in raw_specs.items()}
    _validate(defaults_applied)
    return _to_spec_dataclasses(defaults_applied)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)
import hashlib
from dataclasses import replace
from datetime import datetime, timedelta, timezone
from pathlib import Path

import pytest
from dateutil.parser import parse as parse_date

from scitacean import File, IntegrityError, RemotePath
from scitacean.file import checksum_of_file
from scitacean.logging import logger_name
from scitacean.model import DataFile

from .common.files import make_file


@pytest.fixture
def fake_file(fs):
    return make_file(fs, path="local/dir/events.nxs")


def test_file_from_local(fake_file):
    file = replace(File.from_local(fake_file["path"]), checksum_algorithm="md5")
    assert file.remote_access_path("/remote") is None
    assert file.local_path == fake_file["path"]
    assert file.remote_path == str(fake_file["path"])
    assert file.checksum() == fake_file["checksum"]
    assert file.size == fake_file["size"]
    assert file.remote_uid is None
    assert file.remote_gid is None
    assert file.remote_perm is None
    assert abs(fake_file["creation_time"] - file.creation_time) < timedelta(seconds=1)


def test_file_from_local_with_base_path(fake_file):
    assert str(fake_file["path"]) == "local/dir/events.nxs"  # used below

    file = replace(
        File.from_local(fake_file["path"], base_path="local"), checksum_algorithm="md5"
    )
    assert file.remote_access_path("/remote") is None
    assert file.local_path == fake_file["path"]
    assert file.remote_path == "dir/events.nxs"
    assert file.checksum() == fake_file["checksum"]
    assert file.size == fake_file["size"]
    assert file.remote_uid is None
    assert file.remote_gid is None
    assert file.remote_perm is None
    assert abs(fake_file["creation_time"] - file.creation_time) < timedelta(seconds=1)


def test_file_from_local_set_remote_path(fake_file):
    file = replace(
        File.from_local(fake_file["path"], remote_path="remote/location/file.nxs"),
        checksum_algorithm="md5",
    )
    assert file.remote_access_path("/remote") is None
    assert file.local_path == fake_file["path"]
    assert file.remote_path == "remote/location/file.nxs"
    assert file.checksum() == fake_file["checksum"]
    assert file.size == fake_file["size"]
    assert file.remote_uid is None
    assert file.remote_gid is None
    assert file.remote_perm is None
    assert abs(fake_file["creation_time"] - file.creation_time) < timedelta(seconds=1)


def test_file_from_local_set_many_args(fake_file):
    file = replace(
        File.from_local(
            fake_file["path"],
            base_path="local",
            remote_uid="user-usy",
            remote_gid="groupy-group",
            remote_perm="wrx",
        ),
        checksum_algorithm="md5",
    )
    assert file.remote_access_path("/remote") is None
    assert file.local_path == fake_file["path"]
    assert file.checksum() == fake_file["checksum"]
    assert file.size == fake_file["size"]
    assert file.remote_uid == "user-usy"
    assert file.remote_gid == "groupy-group"
    assert file.remote_perm == "wrx"
    assert abs(fake_file["creation_time"] - file.creation_time) < timedelta(seconds=1)


@pytest.mark.parametrize("alg", ("md5", "sha256", "blake2s"))
def test_file_from_local_select_checksum_algorithm(fake_file, alg):
    file = replace(File.from_local(fake_file["path"]), checksum_algorithm=alg)
    expected = checksum_of_file(fake_file["path"], algorithm=alg)
    assert file.checksum() == expected


def test_file_from_scicat():
    model = DataFile(
        path="dir/image.jpg", size=12345, time=parse_date("2022-06-22T15:42:53.123Z")
    )
    file = File.from_scicat(model)

    assert file.remote_access_path("/remote/folder") == "/remote/folder/dir/image.jpg"
    assert file.local_path is None
    assert file.checksum() is None
    assert file.size == 12345
    assert file.creation_time == parse_date("2022-06-22T15:42:53.123Z")


def test_make_model_local_file(fake_file):
    file = replace(
        File.from_local(
            fake_file["path"],
            remote_uid="user-usy",
            remote_gid="groupy-group",
            remote_perm="wrx",
        ),
        checksum_algorithm="blake2s",
    )
    model = file.make_model()
    assert model.path == str(fake_file["path"])
    assert model.size == fake_file["size"]
    assert model.chk == checksum_of_file(fake_file["path"], algorithm="blake2s")
    assert model.gid == "groupy-group"
    assert model.perm == "wrx"
    assert model.uid == "user-usy"
    assert abs(fake_file["creation_time"] - model.time) < timedelta(seconds=1)


def test_uploaded(fake_file):
    file = replace(
        File.from_local(
            fake_file["path"],
            remote_uid="the-user",
        ),
        checksum_algorithm="sha256",
    )
    uploaded = file.uploaded(
        remote_gid="the-group", remote_creation_time=parse_date("2100-09-07T11:34:51")
    )

    assert uploaded.is_on_local
    assert uploaded.is_on_remote

    assert uploaded.remote_path == str(fake_file["path"])
    assert uploaded.local_path == fake_file["path"]
    assert uploaded.checksum() == checksum_of_file(
        fake_file["path"], algorithm="sha256"
    )
    assert uploaded.remote_uid == "the-user"
    assert uploaded.remote_gid == "the-group"
    assert uploaded._remote_creation_time == parse_date("2100-09-07T11:34:51")


def test_downloaded():
    model = DataFile(
        path="dir/stream.s",
        size=55123,
        time=parse_date("2025-01-09T21:00:21.421Z"),
        perm="xrw",
        createdBy="creator-id",
    )
    file = File.from_scicat(model)
    downloaded = file.downloaded(local_path="/local/stream.s")

    assert downloaded.is_on_local
    assert downloaded.is_on_remote

    assert downloaded.remote_path == RemotePath("dir/stream.s")
    assert downloaded.local_path == Path("/local/stream.s")
    assert downloaded.remote_perm == "xrw"
    assert downloaded.created_by == "creator-id"


def test_creation_time_is_always_local_time(fake_file):
    file = File.from_local(path=fake_file["path"])
    model = file.make_model()
    model.time = parse_date("2105-04-01T04:52:23")
    uploaded = file.uploaded()

    assert abs(fake_file["creation_time"] - uploaded.creation_time) < timedelta(
        seconds=1
    )


def test_size_is_always_local_size(fake_file):
    file = File.from_local(path=fake_file["path"])
    model = file.make_model()
    model.size = 999999999
    uploaded = file.uploaded()

    assert uploaded.size == fake_file["size"]


def test_checksum_is_always_local_checksum(fake_file):
    file = File.from_local(path=fake_file["path"])
    uploaded = replace(
        file.uploaded(), _remote_checksum="6e9eb73953231aebbbc8788f39f08618"
    )

    assert replace(uploaded, checksum_algorithm="md5").checksum() == checksum_of_file(
        fake_file["path"], algorithm="md5"
    )
    assert replace(
        uploaded, checksum_algorithm="sha256"
    ).checksum() == checksum_of_file(fake_file["path"], algorithm="sha256")


def test_creation_time_is_up_to_date(fs, fake_file):
    file = File.from_local(path=fake_file["path"])
    with open(fake_file["path"], "wb") as f:
        f.write(b"some new content to update time stamp")
    new_creation_time = datetime.fromtimestamp(
        fake_file["path"].stat().st_mtime
    ).astimezone(timezone.utc)
    assert file.creation_time == new_creation_time


def test_size_is_up_to_date(fs, fake_file):
    file = File.from_local(path=fake_file["path"])
    new_contents = b"content with a new size"
    assert len(new_contents) != fake_file["size"]
    with open(fake_file["path"], "wb") as f:
        f.write(new_contents)
    assert file.size == len(new_contents)


def test_checksum_is_up_to_date(fs, fake_file):
    file = replace(File.from_local(path=fake_file["path"]), checksum_algorithm="md5")
    new_contents = b"content a different checksum"

    checksum = hashlib.new("md5")
    checksum.update(new_contents)
    checksum = checksum.hexdigest()
    assert checksum != fake_file["size"]

    with open(fake_file["path"], "wb") as f:
        f.write(new_contents)
    assert file.checksum() == checksum


def test_validate_after_download_detects_bad_checksum(fake_file):
    model = DataFile(
        path=fake_file["path"].name,
        size=fake_file["size"],
        time=parse_date("2022-06-22T15:42:53.123Z"),
        chk="incorrect-checksum",
    )
    file = replace(
        File.from_scicat(model),
        checksum_algorithm="md5",
    )
    downloaded = file.downloaded(local_path=fake_file["path"])

    with pytest.raises(IntegrityError):
        downloaded.validate_after_download()


def test_validate_after_download_ignores_checksum_if_no_algorithm(fake_file):
    model = DataFile(
        path=fake_file["path"].name,
        size=fake_file["size"],
        time=parse_date("2022-06-22T15:42:53.123Z"),
        chk="incorrect-checksum",
    )
    file = File.from_scicat(model)
    downloaded = file.downloaded(local_path=fake_file["path"])

    # does not raise
    downloaded.validate_after_download()


def test_validate_after_download_detects_size_mismatch(fake_file, caplog):
    model = DataFile(
        path=fake_file["path"].name,
        size=fake_file["size"] + 100,
        time=parse_date("2022-06-22T15:42:53.123Z"),
        chk="incorrect-checksum",
    )
    file = replace(
        File.from_scicat(model),
        checksum_algorithm=None,
    )
    downloaded = file.downloaded(local_path=fake_file["path"])
    with caplog.at_level("INFO", logger=logger_name()):
        downloaded.validate_after_download()
    assert "does not match size reported in dataset" in caplog.text


@pytest.mark.parametrize("chk", ("sha256", None))
def test_local_is_not_up_to_date_for_remote_file(chk):
    file = File.from_scicat(DataFile(path="data.csv", size=65178, chk=chk))
    assert not file.local_is_up_to_date()


def test_local_is_up_to_date_for_local_file():
    # Note that the file does not actually exist on disk but the test still works.
    file = File.from_local(path="image.jpg")
    assert file.local_is_up_to_date()


def test_local_is_not_up_to_date_without_checksum_alg():
    file = File.from_scicat(
        DataFile(path="data.csv", size=65178, chk="sha256")
    ).downloaded(local_path="data.csv")
    with pytest.warns(UserWarning, match="checksum"):
        assert not file.local_is_up_to_date()


def test_local_is_up_to_date_matching_checksum(fake_file):
    model = DataFile(
        path=fake_file["path"].name,
        size=fake_file["size"],
        time=parse_date("2022-06-22T15:42:53.123Z"),
        chk=fake_file["checksum"],
    )
    file = replace(
        File.from_scicat(model).downloaded(local_path=fake_file["path"]),
        checksum_algorithm="md5",
    )
    assert file.local_is_up_to_date()


def test_local_is_not_up_to_date_differing_checksum(fake_file):
    model = DataFile(
        path=fake_file["path"].name,
        size=fake_file["size"],
        time=parse_date("2022-06-22T15:42:53.123Z"),
        chk="a-different-checksum",
    )
    file = replace(
        File.from_scicat(model).downloaded(local_path=fake_file["path"]),
        checksum_algorithm="md5",
    )
    assert not file.local_is_up_to_date()



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)
from __future__ import annotations

import uuid
from typing import Callable, Generator, Optional, Union


class PID:
    """Stores the ID of database item.

    The ID is split into a prefix and the main identifier.
    The prefix identifies an instance of SciCat and the main identifier a dataset.

    The two components are merged using a "/", i.e.

    .. code-block:: python

        full_id = PID.prefix + '/' + PID.pid

    Equivalently, ``str`` can be used to construct the full id:

    .. code-block:: python

        full_id = str(PID)
    """

    __slots__ = ("_pid", "_prefix")

    def __init__(self, *, pid: str, prefix: Optional[str] = None):
        """Initialize an instance from individual components.

        Parameters
        ----------
        pid:
            Main part of the ID which uniquely identifies a dataset.
        prefix:
            Identifies the instance of SciCat.
        """
        self._pid = pid
        self._prefix = prefix

    @classmethod
    def parse(cls, x: Union[str, PID]) -> PID:
        """Build a PID from a string.

        The string is split at the first "/" to determine
        prefix and main ID.
        This means that it only works if the prefix and main ID do
        not contain any slashes.

        Parameters
        ----------
        x:
            String holding an ID with or without prefix.
            Or a PID object, in which case a copy is returned.

        Returns
        -------
        :
            A new PID object constructed from ``x``.
        """
        if isinstance(x, PID):
            return PID(pid=x.pid, prefix=x.prefix)
        pieces = x.split("/", 1)
        if len(pieces) == 1:
            return PID(pid=pieces[0], prefix=None)
        return PID(prefix=pieces[0], pid=pieces[1])

    @classmethod
    def generate(cls, *, prefix: Optional[str] = None) -> PID:
        """Create a new unique PID.

        Uses UUID4 to generate the ID.

        Parameters
        ----------
        prefix:
            If given, the returned PID has this prefix.

        Returns
        -------
        :
            A new PID object.
        """
        return PID(prefix=prefix, pid=str(uuid.uuid4()))

    @property
    def pid(self) -> str:
        """Main part of the ID."""
        return self._pid

    @property
    def prefix(self) -> Optional[str]:
        """Prefix part of the ID if there is one."""
        return self._prefix

    @property
    def without_prefix(self) -> PID:
        """Return a new PID with the prefix set to None."""
        return PID(pid=self.pid, prefix=None)

    def __str__(self) -> str:
        if self.prefix is not None:
            return self.prefix + "/" + self.pid
        return self.pid

    def __repr__(self) -> str:
        return f"PID(prefix={self.prefix}, pid={self.pid})"

    def __hash__(self) -> int:
        return hash(str(self))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, PID):
            return False
        return self.prefix == other.prefix and self.pid == other.pid

    @classmethod
    def __get_validators__(
        cls,
    ) -> Generator[Callable[[Union[str, PID]], PID], None, None]:
        yield cls.validate

    @classmethod
    def validate(cls, value: Union[str, PID]) -> PID:
        """Pydantic validator for PID fields."""
        if isinstance(value, str):
            return PID.parse(value)
        if isinstance(value, PID):
            return value
        raise TypeError("expected a PID or str")



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)
"""Exception classes."""


class FileUploadError(RuntimeError):
    """Raised when file upload fails."""


class IntegrityError(RuntimeError):
    """Raised when a dataset or file is broken."""


class ScicatCommError(RuntimeError):
    """Raised when communication with SciCat fails."""


class ScicatLoginError(RuntimeError):
    """Raised when login to SciCat server fails."""



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)

from scitacean import Dataset
from scitacean.util.formatter import DatasetPathFormatter


def test_dataset_formatter_uses_dataset_fields():
    dset = Dataset(type="raw", owner="magrat")
    fmt = "{type} {owner}"
    formatted = DatasetPathFormatter().format(fmt, dset)
    expected = fmt.format(type=dset.type, owner=dset.owner)
    assert formatted == expected


def test_dataset_formatter_can_access_attrs_of_fields():
    dset = Dataset(type="raw", pid="prefix/actual-id")
    formatted = DatasetPathFormatter().format("{pid.pid}", dset)
    assert formatted == "actual-id"


def test_dataset_formatter_escapes_characters():
    dset = Dataset(type="raw", owner="Harald Blåtand", owner_email="harald@viking.dk")
    formatted = DatasetPathFormatter().format("{owner}-{owner_email}", dset)
    expected = "Harald Bl_xe5tand-harald_viking.dk"
    assert formatted == expected


def test_dataset_formatter_preserves_path_separators():
    dset = Dataset(type="raw", owner="Nanny Ogg")
    formatted = DatasetPathFormatter().format("{type}/{owner}.data", dset)
    assert formatted == "raw/Nanny Ogg.data"



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)
from functools import partial
from typing import Any, Dict, Optional

from email_validator import EmailNotValidError, validate_email
from hypothesis import strategies as st

from scitacean import Dataset, DatasetType
from scitacean._internal.orcid import orcid_checksum


# email_validator and by extension pydantic is more picky than hypothesis
# so make sure that generated emails actually pass model validation.
def _is_valid_email(email: str) -> bool:
    try:
        validate_email(email, check_deliverability=False)
        return True
    except EmailNotValidError:
        return False


def multi_emails() -> st.SearchStrategy[str]:
    # Convert to lowercase because that is what pydantic does.
    return st.lists(
        st.emails().filter(_is_valid_email).map(lambda s: s.lower()),
        min_size=1,
        max_size=2,
    ).map(lambda email: ";".join(email))


def _email_field_strategy(
    field: Dataset.Field, dataset_type: DatasetType
) -> st.SearchStrategy[Optional[str]]:
    if field.required(dataset_type):
        return multi_emails()
    return st.none() | multi_emails()


def orcids() -> st.SearchStrategy[str]:
    def make_orcid(digits: str) -> str:
        digits = digits[:-1] + orcid_checksum(digits)
        return "https://orcid.org/" + "-".join(
            digits[i : i + 4] for i in range(0, 16, 4)
        )

    return st.text(alphabet="0123456789", min_size=16, max_size=16).map(make_orcid)


def _orcid_field_strategy(
    field: Dataset.Field, dataset_type: DatasetType
) -> st.SearchStrategy[Optional[str]]:
    if field.required(dataset_type):
        return orcids()
    return st.none() | orcids()


def _scientific_metadata_strategy(
    field: Dataset.Field, dataset_type: DatasetType
) -> st.SearchStrategy[Dict[str, Any]]:
    assert field.type == Dict  # noqa: S101 (testing code -> assert is safe)
    return st.dictionaries(
        keys=st.text(),
        values=st.text() | st.dictionaries(keys=st.text(), values=st.text()),
    )


_SPECIAL_FIELDS = {
    "contact_email": _email_field_strategy,
    "owner_email": _email_field_strategy,
    "orcid_of_owner": _orcid_field_strategy,
    "meta": _scientific_metadata_strategy,
}


def _field_strategy(
    field: Dataset.Field, dataset_type: DatasetType
) -> st.SearchStrategy[Any]:
    if (strategy := _SPECIAL_FIELDS.get(field.name)) is not None:
        return strategy(field, dataset_type)

    typ = field.type if field.required(dataset_type) else Optional[field.type]
    return st.from_type(typ)  # type:ignore[arg-type]


def datasets(
    dataset_type: Optional[DatasetType] = None, **fixed: Any
) -> st.SearchStrategy[Dataset]:
    if dataset_type is None:
        return st.sampled_from(DatasetType).flatmap(partial(datasets, **fixed))

    def make_fixed_arg(key: str) -> st.SearchStrategy[Any]:
        val = fixed[key]
        return val if isinstance(val, st.SearchStrategy) else st.just(val)

    def make_arg(field: Dataset.Field) -> st.SearchStrategy[Any]:
        if field.name in fixed:
            return make_fixed_arg(field.name)
        return _field_strategy(
            field, dataset_type=dataset_type  # type: ignore[arg-type]
        )

    def make_args(read_only: bool) -> Dict[str, st.SearchStrategy[Any]]:
        return {
            field.name: make_arg(field)
            for field in Dataset.fields(read_only=read_only, dataset_type=dataset_type)
            if field.name != "type"
        }

    kwargs = make_args(read_only=False)
    read_only_arg = st.fixed_dictionaries(make_args(read_only=True))
    return st.builds(
        Dataset,
        type=st.just(dataset_type),
        _read_only=read_only_arg,
        **kwargs,
    )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)
import hashlib
from datetime import datetime, timedelta, timezone
from pathlib import Path

import pytest
from hypothesis import given
from hypothesis import strategies as st

from scitacean.filesystem import (
    RemotePath,
    checksum_of_file,
    escape_path,
    file_modification_time,
    file_size,
)


def test_remote_path_creation_and_str():
    assert str(RemotePath("/mnt/data/folder/file.txt")) == "/mnt/data/folder/file.txt"
    assert str(RemotePath("dir/image.png")) == "dir/image.png"
    assert str(RemotePath("data.nxs")) == "data.nxs"
    assert str(RemotePath(RemotePath("source/events.h5"))) == "source/events.h5"


def test_remote_path_init_requires_path_like():
    with pytest.raises(TypeError):
        RemotePath(6133)  # type: ignore[arg-type]
    with pytest.raises(TypeError):
        RemotePath(["folder", "file.dat"])  # type: ignore[arg-type]


@pytest.mark.parametrize(
    "types", ((RemotePath, RemotePath), (RemotePath, str), (str, RemotePath))
)
def test_remote_path_eq(types):
    ta, tb = types
    assert ta("/source/data.csv") == tb("/source/data.csv")


@pytest.mark.parametrize(
    "types", ((RemotePath, RemotePath), (RemotePath, str), (str, RemotePath))
)
def test_remote_path_neq(types):
    ta, tb = types
    assert ta("/source/data.csv") != tb("/host/dir/song.mp3")


@pytest.mark.parametrize(
    "types", ((RemotePath, RemotePath), (RemotePath, str), (str, RemotePath))
)
def test_remote_path_join(types):
    ta, tb = types
    assert ta("/source/123") / tb("file.data") == RemotePath("/source/123/file.data")
    assert ta("/source/123/") / tb("file.data") == RemotePath("/source/123/file.data")
    assert ta("/source/123") / tb("/file.data") == RemotePath("/source/123/file.data")
    assert ta("/source/123/") / tb("/file.data") == RemotePath("/source/123/file.data")


@pytest.mark.parametrize(
    "types", ((RemotePath, RemotePath), (RemotePath, str), (str, RemotePath))
)
def test_remote_path_join_url(types):
    ta, tb = types
    assert ta("https://server.eu") / tb("1234-abcd/data.txt") == RemotePath(
        "https://server.eu/1234-abcd/data.txt"
    )
    assert ta("https://server.eu/") / tb("1234-abcd/data.txt") == RemotePath(
        "https://server.eu/1234-abcd/data.txt"
    )
    assert ta("https://server.eu") / tb("/1234-abcd/data.txt") == RemotePath(
        "https://server.eu/1234-abcd/data.txt"
    )
    assert ta("https://server.eu/") / tb("/1234-abcd/data.txt") == RemotePath(
        "https://server.eu/1234-abcd/data.txt"
    )


def test_remote_path_name():
    assert RemotePath("table.csv").name == "table.csv"
    assert RemotePath("README").name == "README"
    assert RemotePath("path/").name == "path"
    assert RemotePath("dir/folder/file1.txt").name == "file1.txt"


def test_remote_path_suffix():
    assert RemotePath("file.txt").suffix == ".txt"
    assert RemotePath("folder/image.png").suffix == ".png"
    assert RemotePath("dir/table.txt.csv").suffix == ".csv"
    assert RemotePath("archive.tz/data.dat").suffix == ".dat"
    assert RemotePath("location.dir/file").suffix is None
    assert RemotePath("source/file").suffix is None


def test_remote_path_truncated():
    assert RemotePath("something-long.txt").truncated(10) == "someth.txt"
    assert RemotePath("longlonglong/short").truncated(5) == "longl/short"
    assert RemotePath("a-long.data.dir/filename.csv").truncated(7) == "a-l.dir/fil.csv"
    assert RemotePath("file.longextension").truncated(9) == "f.longext"


@pytest.mark.parametrize("size", (0, 1, 57121))
def test_file_size(fs, size):
    fs.create_file("image.tiff", st_size=size)
    assert file_size(Path("image.tiff")) == size


def test_file_modification_time(fs):
    fs.create_file("data.dat")
    expected = datetime.now(tz=timezone.utc)
    assert abs(file_modification_time(Path("data.dat")) - expected) < timedelta(
        seconds=5
    )


@pytest.mark.parametrize(
    "contents",
    (b"small file contents", b"large contents " * 100000),
    ids=("small", "large"),
)
def test_checksum_of_file(fs, contents):
    fs.create_file("file.txt", contents=contents)

    assert (
        checksum_of_file("file.txt", algorithm="md5")
        == hashlib.md5(contents).hexdigest()
    )
    assert (
        checksum_of_file("file.txt", algorithm="sha256")
        == hashlib.sha256(contents).hexdigest()
    )


@pytest.mark.parametrize("path_type", (str, Path, RemotePath))
def test_escape_path_returns_same_type_as_input(path_type):
    assert isinstance(escape_path(path_type("x")), path_type)


@given(st.text())
def test_escape_path_returns_ascii(path):
    # does not raise
    escape_path(path).encode("ascii", "strict")


@given(st.text())
def test_escape_path_returns_no_bad_character(path):
    bad_linux = "/" + chr(0)
    bad_windows = '<>:"\\/|?*' + "".join(map(chr, range(0, 31)))
    escaped = escape_path(path)
    assert not any(c in escaped for c in bad_linux + bad_windows)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)
"""Generate pydantic models."""

import dataclasses
from graphlib import TopologicalSorter
from pathlib import Path
from typing import Dict, List

from spec import Field, Spec
from templates import BANNER, load_template
from util import quote


def _write_model(target: Path, models: list):
    with target.open("w") as f:
        f.write(
            BANNER + load_template("model").substitute(models="\n\n\n".join(models))
        )


def _format_model_field(field: Field) -> str:
    name = field.model_name
    typ = field.type if field.required else f"Optional[{field.type}]"
    return f"    {name}: {typ}"


def _format_validator(validator: str, fields: List[Field]) -> str:
    to_validate = [
        field.model_name for field in fields if field.validation == validator
    ]
    if not to_validate:
        return ""
    return f"""    @pydantic.validator({", ".join(map(quote, to_validate))})
    def _validate_{validator}(cls, value: Any) -> Any:
        return _validate_{validator}(value)"""


def _format_validators(fields: List[Field]) -> str:
    res = "\n\n".join(
        (
            _format_validator("emails", fields),
            _format_validator("size", fields),
            _format_validator("orcid", fields),
        )
    )
    return "\n    " + res.strip()


def _format_model(spec: Spec) -> str:
    head = f"""class {spec.name}({spec.inherits}):\n"""
    fields = sorted(
        sorted(spec.fields, key=lambda f: f.name), key=lambda f: not f.required
    )
    attributes = "\n".join(_format_model_field(field) for field in fields)
    validations = _format_validators(fields)
    return head + attributes + ("\n" + validations if validations.strip() else "")


def _make_dataset_type(spec: Spec, typ: str) -> Spec:
    def handle_field(field: Field) -> Field:
        return dataclasses.replace(
            field,
            model_name=field.model_name[typ]
            if isinstance(field.model_name, dict)
            else field.model_name,
        )

    return Spec(
        name="DerivedDataset" if typ == "derived" else "RawDataset",
        inherits=spec.inherits,
        fields=[
            handle_field(field) for field in spec.fields if typ in field.extra["used"]
        ],
    )


def _split_dataset(specs: Dict[str, Spec]) -> Dict[str, Spec]:
    specs = dict(specs)
    dataset = specs.pop("Dataset")
    specs["DerivedDataset"] = _make_dataset_type(dataset, "derived")
    specs["RawDataset"] = _make_dataset_type(dataset, "raw")
    return specs


def generate_models(target: Path, specs: Dict[str, Spec]):
    specs = _split_dataset(specs)
    spec_order = TopologicalSorter(
        {
            key: {val.inherits}
            for key, val in sorted(specs.items(), key=lambda kv: kv[0])
        },
    ).static_order()
    _write_model(
        target,
        [_format_model(specs[name]) for name in spec_order if name != "BaseModel"],
    )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)
from dateutil.parser import parse as parse_datetime
from hypothesis import given, settings
from hypothesis import strategies as st

from scitacean import PID, DatasetType
from scitacean.testing import strategies as sst


@given(sst.datasets())
def test_datasets_makes_valid_dataset(dset):
    _ = dset.make_model()


@settings(max_examples=10)
@given(sst.datasets(dataset_type=DatasetType.RAW))
def test_datasets_can_set_type_to_raw(dset):
    assert dset.type == "raw"


@settings(max_examples=10)
@given(sst.datasets(dataset_type=DatasetType.DERIVED))
def test_datasets_can_set_type_to_derived(dset):
    assert dset.type == "derived"


@settings(max_examples=20)
@given(
    sst.datasets(
        name="my-dataset-name",
        access_groups=["group1", "another-group"],
        source_folder="/the/source/of/truth",
        created_at=parse_datetime("2123-11-29T17:44:56Z"),
        pid=PID(prefix="prefix", pid="some-id"),
        description=None,
    )
)
def test_datasets_can_fix_fields(dset):
    assert dset.name == "my-dataset-name"
    assert dset.access_groups == ["group1", "another-group"]
    assert dset.source_folder == "/the/source/of/truth"
    assert dset.created_at == parse_datetime("2123-11-29T17:44:56Z")
    assert dset.pid == PID(prefix="prefix", pid="some-id")
    assert dset.description is None


@settings(max_examples=20)
@given(
    sst.datasets(
        created_by=st.just("the-creator"),
        owner=st.sampled_from(("owner1", "owner2")),
        name=st.text(min_size=2, max_size=5),
    )
)
def test_datasets_can_set_strategy_for_fields(dset):
    assert dset.created_by == "the-creator"
    assert dset.owner in ("owner1", "owner2")
    assert 2 <= len(dset.name) <= 5



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)
"""Common utilities for file transfers."""

from typing import Optional, Union

from ..dataset import Dataset
from ..filesystem import RemotePath
from ..util.formatter import DatasetPathFormatter


def source_folder_for(
    dataset: Dataset, pattern: Optional[Union[str, RemotePath]]
) -> RemotePath:
    """Get or build the source folder for a dataset.

    Parameters
    ----------
    dataset:
        Build the source folder for this dataset.
    pattern:
        A string for constructing a source folder.
        Can have placeholders to be used by ``DatasetPathFormatter``.

    Returns
    -------
    :
        The source folder for the dataset.
    """
    if pattern is None:
        if dataset.source_folder is None:
            raise ValueError(
                "Cannot determine source_folder for dataset. "
                "Either the dataset's source_folder or the "
                "file transfer's source_folder must be set."
            )
        return dataset.source_folder

    return RemotePath(DatasetPathFormatter().format(str(pattern), dataset))



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)

from datetime import datetime, timedelta, timezone
from pathlib import Path

import pytest
from hypothesis import assume, given, settings
from hypothesis import strategies as st

from scitacean import PID, Dataset, DatasetType, File, model
from scitacean.testing import strategies as sst

from .common.files import make_file


@pytest.mark.parametrize("typ", (DatasetType.RAW, DatasetType.DERIVED))
def test_new_dataset_has_no_files(typ):
    dset = Dataset(type=typ)
    assert len(list(dset.files)) == 0
    assert dset.number_of_files == 0
    assert dset.number_of_files_archived == 0
    assert dset.packed_size == 0
    assert dset.size == 0


@pytest.mark.parametrize("typ", (DatasetType.RAW, DatasetType.DERIVED))
def test_add_local_file_to_new_dataset(typ, fs):
    file_data = make_file(fs, "local/folder/data.dat")

    dset = Dataset(type=typ)
    dset.add_local_files("local/folder/data.dat")

    assert len(list(dset.files)) == 1
    assert dset.number_of_files == 1
    assert dset.number_of_files_archived == 0
    assert dset.packed_size == 0
    assert dset.size == file_data["size"]

    [f] = dset.files
    assert not f.is_on_remote
    assert f.is_on_local
    assert f.remote_access_path(dset.source_folder) is None
    assert f.local_path == Path("local/folder/data.dat")
    assert f.size == file_data["size"]
    assert f.checksum_algorithm == "md5"

    assert abs(file_data["creation_time"] - f.creation_time) < timedelta(seconds=1)
    assert abs(file_data["creation_time"] - f.make_model().time) < timedelta(seconds=1)


@pytest.mark.parametrize("typ", (DatasetType.RAW, DatasetType.DERIVED))
def test_add_multiple_local_files_to_new_dataset(typ, fs):
    file_data0 = make_file(fs, "common/location1/data.dat")
    file_data1 = make_file(fs, "common/song.mp3")

    dset = Dataset(type=typ)
    dset.add_local_files("common/location1/data.dat", "common/song.mp3")

    assert len(list(dset.files)) == 2
    assert dset.number_of_files == 2
    assert dset.number_of_files_archived == 0
    assert dset.packed_size == 0
    assert dset.size == file_data0["size"] + file_data1["size"]

    [f0, f1] = dset.files
    if f0.local_path.suffix == ".mp3":
        f1, f0 = f0, f1

    assert not f0.is_on_remote
    assert f0.is_on_local
    assert f0.remote_access_path(dset.source_folder) is None
    assert f0.local_path == Path("common/location1/data.dat")
    assert f0.size == file_data0["size"]
    assert f0.make_model().path == "common/location1/data.dat"
    assert f0.checksum_algorithm == "md5"

    assert not f1.is_on_remote
    assert f1.is_on_local
    assert f1.remote_access_path(dset.source_folder) is None
    assert f1.local_path == Path("common/song.mp3")
    assert f1.size == file_data1["size"]
    assert f1.make_model().path == "common/song.mp3"
    assert f1.checksum_algorithm == "md5"


@pytest.mark.parametrize("typ", (DatasetType.RAW, DatasetType.DERIVED))
def test_add_multiple_local_files_to_new_dataset_with_base_path(typ, fs):
    file_data0 = make_file(fs, "common/location1/data.dat")
    file_data1 = make_file(fs, "common/song.mp3")

    dset = Dataset(type=typ)
    dset.add_local_files(
        "common/location1/data.dat", "common/song.mp3", base_path="common"
    )

    assert len(list(dset.files)) == 2
    assert dset.number_of_files == 2
    assert dset.number_of_files_archived == 0
    assert dset.packed_size == 0
    assert dset.size == file_data0["size"] + file_data1["size"]

    [f0, f1] = dset.files
    if f0.local_path.suffix == ".mp3":
        f1, f0 = f0, f1

    assert not f0.is_on_remote
    assert f0.is_on_local
    assert f0.remote_access_path(dset.source_folder) is None
    assert f0.local_path == Path("common/location1/data.dat")
    assert f0.size == file_data0["size"]
    assert f0.make_model().path == "location1/data.dat"
    assert f0.checksum_algorithm == "md5"

    assert not f1.is_on_remote
    assert f1.is_on_local
    assert f1.remote_access_path(dset.source_folder) is None
    assert f1.local_path == Path("common/song.mp3")
    assert f1.size == file_data1["size"]
    assert f1.make_model().path == "song.mp3"
    assert f1.checksum_algorithm == "md5"


@pytest.mark.parametrize("typ", (DatasetType.RAW, DatasetType.DERIVED))
@pytest.mark.parametrize("algorithm", ("sha256", None))
def test_can_set_default_checksum_algorithm(typ, algorithm, fs):
    make_file(fs, "local/data.dat")

    dset = Dataset(type=typ, checksum_algorithm=algorithm)
    dset.add_local_files("local/data.dat")

    [f] = dset.files
    assert f.checksum_algorithm == algorithm


@given(sst.datasets())
@settings(max_examples=100)
def test_dataset_models_roundtrip(initial):
    dataset_model = initial.make_model()
    dblock_models = initial.make_datablock_models()
    rebuilt = Dataset.from_models(
        dataset_model=dataset_model, orig_datablock_models=dblock_models.orig_datablocks
    )
    assert initial == rebuilt


@given(sst.datasets())
@settings(max_examples=10)
def test_make_scicat_models_datablock_without_files(dataset):
    assert dataset.make_datablock_models().orig_datablocks is None


@given(sst.datasets())
@settings(max_examples=10)
def test_make_scicat_models_datablock_with_one_file(dataset):
    file_model = model.DataFile(path="path", size=6163, chk="8450ac0", gid="group")
    dataset.add_files(File.from_scicat(local_path=None, model=file_model))

    blocks = dataset.make_datablock_models().orig_datablocks
    assert len(blocks) == 1

    block = blocks[0]
    assert block.size == 6163
    assert block.datasetId == dataset.pid
    assert block.dataFileList == [file_model]


@given(sst.datasets())
@settings(max_examples=10)
def test_eq_self(dset):
    dset.add_files(
        File.from_scicat(
            local_path=None,
            model=model.DataFile(path="path", size=94571),
        )
    )
    assert dset == dset


# Filtering out bools because hypothesis struggles to find enough examples
# where the new value is not the same as the old value.
@pytest.mark.parametrize(
    "field",
    filter(lambda f: f.type != bool, Dataset.fields(read_only=False)),
    ids=lambda f: f.name,
)
@given(sst.datasets(), st.data())
@settings(max_examples=10)
def test_neq_single_mismatched_field_writable(field, initial, data):
    new_val = data.draw(st.from_type(field.type))
    assume(new_val != getattr(initial, field.name))
    modified = initial.replace(**{field.name: new_val})
    assert modified != initial


@given(sst.datasets())
@settings(max_examples=10)
def test_neq_single_mismatched_file(initial):
    modified = initial.replace()
    modified.add_files(
        File.from_scicat(
            local_path=None,
            model=model.DataFile(path="path", size=51553312),
        )
    )
    initial.add_files(
        File.from_scicat(
            local_path=None,
            model=model.DataFile(path="path", size=94571),
        )
    )
    assert modified != initial


@given(sst.datasets())
@settings(max_examples=10)
def test_neq_extra_file(initial):
    modified = initial.replace()
    modified.add_files(
        File.from_scicat(
            local_path="/local",
            model=model.DataFile(path="path", size=51553312),
        )
    )
    assert modified != initial


@pytest.mark.parametrize(
    "field",
    Dataset.fields(read_only=False),
    ids=lambda f: f.name,
)
@given(sst.datasets(), st.data())
@settings(max_examples=5)
def test_replace_replaces_single_writable_field(field, initial, data):
    val = data.draw(st.from_type(field.type))
    replaced = initial.replace(**{field.name: val})
    assert getattr(replaced, field.name) == val


@pytest.mark.parametrize(
    "field",
    filter(
        lambda f: f.name
        not in ("number_of_files", "number_of_files_archived", "size", "packed_size"),
        Dataset.fields(read_only=True),
    ),
    ids=lambda f: f.name,
)
@given(sst.datasets(), st.data())
@settings(max_examples=5)
def test_replace_replaces_single_read_only_field(field, initial, data):
    val = data.draw(st.from_type(field.type))
    replaced = initial.replace(_read_only={field.name: val})
    assert getattr(replaced, field.name) == val


@given(sst.datasets())
@settings(max_examples=5)
def test_replace_replaces_multiple_fields(initial):
    replaced = initial.replace(
        owner="a-new-owner",
        used_software=["software1"],
        _read_only={"created_by": "the-creator"},
    )
    assert replaced.owner == "a-new-owner"
    assert replaced.used_software == ["software1"]
    assert replaced.created_by == "the-creator"


@given(sst.datasets())
@settings(max_examples=5)
def test_replace_other_fields_are_copied(initial):
    replaced = initial.replace(
        investigator="inv@esti.gator",
        techniques=[model.Technique(pid="tech/abcd.01", name="magick")],
        type="raw",
        _read_only={"version": 666},
    )
    assert replaced.owner == initial.owner
    assert replaced.size == initial.size
    assert replaced.updated_at == initial.updated_at
    assert replaced.history == initial.history


@given(sst.datasets())
@settings(max_examples=1)
def test_replace_rejects_bad_arguments(initial):
    with pytest.raises(TypeError):
        initial.replace(this_is_not_a_valid="argument", owner="the-owner-of-it-all")


@given(sst.datasets())
@settings(max_examples=1)
def test_replace_does_not_change_files_no_input_files(initial):
    replaced = initial.replace(owner="a-new-owner")
    assert replaced.number_of_files == 0
    assert replaced.size == 0
    assert list(replaced.files) == []


@given(sst.datasets())
@settings(max_examples=1)
def test_replace_does_not_change_files_with_input_files(initial):
    file = File.from_scicat(
        local_path=None,
        model=model.DataFile(path="path", size=6163),
    )
    initial.add_files(file)
    replaced = initial.replace(owner="a-new-owner")
    assert replaced.number_of_files == 1
    assert replaced.size == 6163
    assert list(replaced.files) == list(initial.files)


@given(sst.datasets())
@settings(max_examples=5)
def test_as_new(initial):
    new = initial.as_new()
    assert new.created_at is None
    assert new.created_by is None
    assert new.updated_at is None
    assert new.updated_by is None
    assert new.history == []
    assert new.lifecycle is None
    assert abs(new.creation_time - datetime.now(tz=timezone.utc)) < timedelta(seconds=1)

    assert new.number_of_files == initial.number_of_files
    assert new.size == initial.size
    assert new.name == initial.name
    assert new.input_datasets == initial.input_datasets
    assert new.data_format == initial.data_format


@given(sst.datasets(pid=PID(pid="some-id")))
@settings(max_examples=5)
def test_derive_default(initial):
    derived = initial.derive()
    assert derived.type == "derived"
    assert derived.input_datasets == [initial.pid]
    assert derived.lifecycle is None
    assert derived.history == []

    assert derived.investigator == initial.investigator
    assert derived.owner == initial.owner
    assert derived.orcid_of_owner == initial.orcid_of_owner
    assert derived.owner_email == initial.owner_email
    assert derived.contact_email == initial.contact_email
    assert derived.techniques == initial.techniques

    assert derived.name is None
    assert derived.used_software is None
    assert derived.number_of_files == 0
    assert derived.number_of_files_archived == 0


@given(sst.datasets(pid=PID(pid="some-id")))
@settings(max_examples=5)
def test_derive_set_keep(initial):
    derived = initial.derive(keep=("name", "used_software"))
    assert derived.type == "derived"
    assert derived.input_datasets == [initial.pid]
    assert derived.lifecycle is None
    assert derived.history == []

    assert derived.name == initial.name
    assert derived.used_software == initial.used_software

    assert derived.number_of_files == 0
    assert derived.number_of_files_archived == 0


@given(sst.datasets(pid=PID(pid="some-id")))
@settings(max_examples=5)
def test_derive_keep_nothing(initial):
    derived = initial.derive(keep=())
    assert derived.type == "derived"
    assert derived.input_datasets == [initial.pid]
    assert derived.lifecycle is None
    assert derived.history == []

    assert derived.investigator is None
    assert derived.owner is None
    assert derived.name is None
    assert derived.used_software is None
    assert derived.number_of_files == 0
    assert derived.number_of_files_archived == 0


@given(sst.datasets(pid=None))
@settings(max_examples=5)
def test_derive_requires_pid(initial):
    with pytest.raises(ValueError):
        initial.derive()



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)

import os
from contextlib import contextmanager
from datetime import datetime, tzinfo
from getpass import getpass
from pathlib import Path
from typing import Any, Iterator, List, Optional, Tuple, Union

from dateutil.tz import gettz

# Note that invoke and paramiko are dependencies of fabric.
from fabric import Connection
from invoke.exceptions import UnexpectedExit
from paramiko import SFTPClient
from paramiko.ssh_exception import AuthenticationException, PasswordRequiredException

from ..dataset import Dataset
from ..error import FileUploadError
from ..file import File
from ..filesystem import RemotePath
from ..logging import get_logger
from .util import source_folder_for

# TODO process multiple files together
# TODO pass pid in put/revert?
#      downloading does not need a pid, so it should not be required in the constructor/


class SSHDownloadConnection:
    def __init__(self, *, connection: Connection) -> None:
        self._connection = connection

    def download_files(self, *, remote: List[RemotePath], local: List[Path]) -> None:
        """Download files from the given remote path."""
        for r, l in zip(remote, local):
            self.download_file(remote=r, local=l)

    def download_file(self, *, remote: RemotePath, local: Path) -> None:
        get_logger().info(
            "Downloading file %s from host %s to %s",
            remote,
            self._connection.host,
            local,
        )
        self._connection.get(remote=str(remote), local=str(local))


class SSHUploadConnection:
    def __init__(self, *, connection: Connection, source_folder: RemotePath) -> None:
        self._connection = connection
        self._source_folder = source_folder
        self._remote_timezone = self._get_remote_timezone()

    @property
    def _sftp(self) -> SFTPClient:
        return self._connection.sftp()  # type: ignore[no-any-return]

    @property
    def source_folder(self) -> RemotePath:
        return self._source_folder

    def remote_path(self, filename: Union[str, RemotePath]) -> RemotePath:
        return self.source_folder / filename

    def _make_source_folder(self) -> None:
        try:
            self._connection.run(f"mkdir -p {os.fspath(self.source_folder)}", hide=True)
        except OSError as exc:
            raise FileUploadError(
                f"Failed to create source folder {self.source_folder}: {exc.args}"
            ) from None

    def upload_files(self, *files: File) -> List[File]:
        """Upload files to the remote folder."""
        self._make_source_folder()
        uploaded = []
        try:
            for file in files:
                uploaded.append(self._upload_file(file))
        except Exception:
            self.revert_upload(*uploaded)
            raise
        return uploaded

    def _upload_file(self, file: File) -> File:
        remote_path = self.remote_path(file.remote_path)
        get_logger().info(
            "Uploading file %s to %s on host %s",
            file.local_path,
            remote_path,
            self._connection.host,
        )
        self._sftp.put(
            remotepath=os.fspath(remote_path), localpath=os.fspath(file.local_path)
        )
        st = self._sftp.stat(os.fspath(remote_path))
        self._validate_upload(file)
        creation_time = (
            datetime.fromtimestamp(st.st_mtime, tz=self._remote_timezone)
            if st.st_mtime
            else None
        )
        return file.uploaded(
            remote_gid=str(st.st_gid),
            remote_uid=str(st.st_uid),
            remote_creation_time=creation_time,
            remote_perm=str(st.st_mode),
            remote_size=st.st_size,
        )

    def _validate_upload(self, file: File) -> None:
        if (checksum := self._compute_checksum(file)) is None:
            return
        if checksum != file.checksum():
            raise FileUploadError(
                f"Upload of file {file.remote_path} failed: "
                f"Checksum of uploaded file ({checksum}) does not "
                f"match checksum of local file ({file.checksum()}) "
                f"using algorithm {file.checksum_algorithm}"
            )

    def _compute_checksum(self, file: File) -> Optional[str]:
        if (hash_exe := _coreutils_checksum_for(file)) is None:
            return None
        try:
            res = self._connection.run(
                f"{hash_exe} {self.remote_path(file.remote_path)}", hide=True
            )
        except UnexpectedExit as exc:
            if exc.result.return_code == 127:
                get_logger().warning(
                    "Cannot validate checksum of uploaded file %s because checksum "
                    "algorithm '%s' is not implemented on the server.",
                    file.remote_path,
                    file.checksum_algorithm,
                )
                return None
            raise
        return res.stdout.split(" ", 1)[0]  # type: ignore[no-any-return]

    def _get_remote_timezone(self) -> tzinfo:
        cmd = 'date +"%Z"'
        tz_str = self._connection.run(cmd, hide=True).stdout.strip()
        if (tz := gettz(tz_str)) is not None:
            return tz
        raise RuntimeError(
            "Failed to get timezone of remote fileserver. "
            f"Command {cmd} returned '{tz_str}' which "
            "cannot be parsed as a timezone."
        )

    def revert_upload(self, *files: File) -> None:
        """Remove uploaded files from the remote folder."""
        for file in files:
            self._revert_upload_single(
                remote=file.remote_path, local=file.local_path
            )  # type: ignore[arg-type]

        if _folder_is_empty(self._connection, self.source_folder):
            try:
                get_logger().info(
                    "Removing empty remote directory %s on host %s",
                    self.source_folder,
                    self._connection.host,
                )
                self._sftp.rmdir(os.fspath(self.source_folder))
            except UnexpectedExit as exc:
                get_logger().warning(
                    "Failed to remove empty remote directory %s on host:\n%s",
                    self.source_folder,
                    self._connection.host,
                    exc.result,
                )

    def _revert_upload_single(self, *, remote: RemotePath, local: Path) -> None:
        remote_path = self.remote_path(remote)
        get_logger().info(
            "Reverting upload of file %s to %s on host %s",
            local,
            remote_path,
            self._connection.host,
        )

        try:
            self._sftp.remove(os.fspath(remote_path))
        except UnexpectedExit as exc:
            get_logger().warning(
                "Error reverting file %s:\n%s", remote_path, exc.result
            )
            return


class SSHFileTransfer:
    """Upload / download files using SSH.

    Configuration & Authentication
    ------------------------------
    The file transfer connects to the server at the address given
    as the ``host`` constructor argument.
    This may be

    - a full url such as ``some.fileserver.edu``,
    - an IP address like ``127.0.0.1``,
    - or a host defined in the user's openSSH config file.

    The file transfer can authenticate using username+password.
    It will ask for those on the command line.
    However, it is **highly recommended to set up a key and use an SSH agent!**
    This increases security as Scitacean no longer has to handle credentials itself.
    And it is required for automated programs where a user cannot enter credentials
    on a command line.

    Upload folder
    -------------
    The file transfer can take an optional ``source_folder`` as a constructor argument.
    If it is given, ``SSHFileTransfer`` uploads all files to it and ignores the
    source folder set in the dataset.
    If it is not given, ``SSHFileTransfer`` uses the dataset's source folder.

    The source folder argument to ``SSHFileTransfer`` may be a Python format string.
    In that case, all format fields are replaced by the corresponding fields
    of the dataset.
    All non-ASCII characters and most special ASCII characters are replaced.
    This should avoid broken paths from essentially random contents in datasets.

    Examples
    --------
    Given

    .. code-block:: python

        dset = Dataset(type="raw",
                       name="my-dataset",
                       source_folder="/dataset/source",
                       )

    This uploads to ``/dataset/source``:

    .. code-block:: python

        file_transfer = SSHFileTransfer(host="fileserver")

    This uploads to ``/transfer/folder``:

    .. code-block:: python

        file_transfer = SSHFileTransfer(host="fileserver",
                                        source_folder="transfer/folder")

    This uploads to ``/transfer/my-dataset``:
    (Note that ``{name}`` is replaced by ``dset.name``.)

    .. code-block:: python

        file_transfer = SSHFileTransfer(host="fileserver",
                                        source_folder="transfer/{name}")

    A useful approach is to include the PID in the source folder, for example
    ``/some/base/folder/{pid.pid}``, to avoid clashes between different datasets.
    """

    def __init__(
        self,
        *,
        host: str,
        port: Optional[int] = None,
        source_folder: Optional[Union[str, Path]] = None,
    ) -> None:
        """Construct a new SSH file transfer.

        Parameters
        ----------
        host:
            URL or name of the server to connect to.
        port:
            Port of the server.
        source_folder:
            Upload files to this folder if set.
            Otherwise, upload to the dataset's source_folder.
            Ignored when downloading files.
        """
        self._host = host
        self._port = port
        self._source_folder_pattern = (
            RemotePath(source_folder) if source_folder is not None else None
        )

    def source_folder_for(self, dataset: Dataset) -> RemotePath:
        """Return the source folder used for the given dataset."""
        return source_folder_for(dataset, self._source_folder_pattern)

    @contextmanager
    def connect_for_download(self) -> Iterator[SSHDownloadConnection]:
        """Create a connection for downloads, use as a context manager."""
        con = _connect(self._host, self._port)
        try:
            yield SSHDownloadConnection(connection=con)
        finally:
            con.close()

    @contextmanager
    def connect_for_upload(self, dataset: Dataset) -> Iterator[SSHUploadConnection]:
        """Create a connection for uploads, use as a context manager."""
        source_folder = self.source_folder_for(dataset)
        con = _connect(self._host, self._port)
        try:
            yield SSHUploadConnection(
                connection=con,
                source_folder=source_folder,
            )
        finally:
            con.close()


def _ask_for_key_passphrase() -> str:
    return getpass("The private key is encrypted, enter passphrase: ")


def _ask_for_credentials(host: str) -> Tuple[str, str]:
    print(f"You need to authenticate to access {host}")  # noqa: T201
    username = input("Username: ")
    password = getpass("Password: ")
    return username, password


def _generic_connect(host: str, port: Optional[int], **kwargs: Any) -> Connection:
    con = Connection(host=host, port=port, **kwargs)
    con.open()
    return con


def _unauthenticated_connect(host: str, port: Optional[int]) -> Connection:
    return _generic_connect(host=host, port=port)


def _authenticated_connect(
    host: str, port: Optional[int], exc: AuthenticationException
) -> Connection:
    # TODO fail fast if output going to file
    if isinstance(exc, PasswordRequiredException) and "encrypted" in exc.args[0]:
        # TODO does not work anymore, exception is always AuthenticationException
        return _generic_connect(
            host=host,
            port=port,
            connect_kwargs={"passphrase": _ask_for_key_passphrase()},
        )
    else:
        username, password = _ask_for_credentials(host)
        return _generic_connect(
            host=host, port=port, user=username, connect_kwargs={"password": password}
        )


def _connect(host: str, port: Optional[int]) -> Connection:
    try:
        try:
            return _unauthenticated_connect(host, port)
        except AuthenticationException as exc:
            return _authenticated_connect(host, port, exc)
    except Exception as exc:
        # We pass secrets as arguments to functions called in this block and those
        # can be leaked through exception handlers. So catch all exceptions
        # and strip the backtrace up to this point to hide those secrets.
        raise type(exc)(exc.args) from None
    except BaseException as exc:
        raise type(exc)(exc.args) from None


def _folder_is_empty(con: Connection, path: RemotePath) -> bool:
    try:
        ls: str = con.run(f"ls {path}", hide=True).stdout
        return ls == ""
    except UnexpectedExit:
        return False  # no further processing is needed in this case


def _coreutils_checksum_for(file: File) -> Optional[str]:
    # blake2s is not supported because `b2sum -l 256` produces a different digest
    # and I don't know why.
    supported = {
        "md5": "md5sum -b",
        "sha256": "sha256sum -b",
        "sha384": "sha384sum -b",
        "sha512": "sha512sum -b",
        "blake2b": "b2sum -l 512 -b",
    }
    algorithm = file.checksum_algorithm
    if algorithm == "blake2s" or algorithm not in supported:
        get_logger().warning(
            "Cannot validate checksum of uploaded file %s because checksum algorithm "
            "'%s' is not supported by scitacean for remote files.",
            file.remote_path,
            file.checksum_algorithm,
        )
        return None
    return supported[algorithm]



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)
import pickle
from copy import deepcopy

import pydantic
import pytest
from dateutil.parser import parse as parse_date

from scitacean import PID, Client, ScicatCommError
from scitacean.model import DataFile, DatasetType, DerivedDataset, OrigDatablock
from scitacean.testing.client import FakeClient
from scitacean.util.credentials import SecretStr

from . import data
from .common.backend import skip_if_not_backend


@pytest.fixture
def derived_dataset():
    return data.as_dataset_model(data.load_datasets()[0])


@pytest.fixture
def orig_datablock():
    return data.as_orig_datablock_model(data.load_orig_datablocks()[0])


def make_fake_client(dataset, datablock):
    client = FakeClient(file_transfer=None)
    client.datasets[dataset.pid] = dataset
    client.orig_datablocks[dataset.pid] = [datablock]
    return client


@pytest.fixture(params=["real", "fake"])
def client(
    request,
    derived_dataset,
    orig_datablock,
    scicat_access,
    scicat_backend,
):
    if request.param == "fake":
        return make_fake_client(derived_dataset, orig_datablock)
    if request.param == "real":
        skip_if_not_backend(request)
        return Client.from_credentials(
            url=scicat_access.url, **scicat_access.functional_credentials
        )


@pytest.fixture
def scicat_client(client):
    return client.scicat


def test_from_token_fake():
    # This should not call the API
    client = FakeClient.from_token(url="some.url/api/v3", token="a-token")  # noqa: S106
    assert isinstance(client, FakeClient)


def test_from_credentials_fake():
    # This should not call the API
    client = FakeClient.from_credentials(
        url="some.url/api/v3",
        username="someone",
        password="the-fake-does-not-care",  # noqa: S106
    )
    assert isinstance(
        client,
        FakeClient,
    )


def test_from_credentials_real(scicat_access, scicat_backend):
    if not scicat_backend:
        pytest.skip("No backend")
    Client.from_credentials(
        url=scicat_access.url, **scicat_access.functional_credentials
    )


def test_get_dataset_model(scicat_client, derived_dataset):
    dset = scicat_client.get_dataset_model(derived_dataset.pid)
    assert dset == derived_dataset


def test_get_dataset_model_bad_id(scicat_client):
    with pytest.raises(ScicatCommError):
        scicat_client.get_dataset_model("bad-pid")


def test_get_orig_datablock(scicat_client, orig_datablock):
    dblock = scicat_client.get_orig_datablocks(orig_datablock.datasetId)
    assert [orig_datablock] == dblock


def test_get_orig_datablock_bad_id(scicat_client):
    with pytest.raises(ScicatCommError):
        scicat_client.get_orig_datablocks("bollocks")


@pytest.mark.parametrize("pid", ("PID.SAMPLE.PREFIX/jeck", "kamelle", None))
def test_create_dataset_model(pid, scicat_client):
    dset = DerivedDataset(
        pid=pid,
        contactEmail="black.foess@dom.koelle",
        creationTime=parse_date("1995-11-11T11:11:11.000Z"),
        owner="bfoess",
        investigator="b.foess@dom.koelle",
        sourceFolder="/dom/platt",
        type=DatasetType.DERIVED,
        inputDatasets=[],
        usedSoftware=[],
        ownerGroup="bfoess",
        accessGroups=["koelle"],
    )
    finalized = scicat_client.create_dataset_model(dset)
    downloaded = scicat_client.get_dataset_model(finalized.pid)
    for key, expected in finalized.dict(exclude_none=True).items():
        # The database populates a number of fields that are None in dset.
        # But we don't want to test those here as we don't want to test the database.
        assert expected == downloaded.dict(exclude_none=True)[key], f"key = {key}"


def test_create_dataset_model_id_clash(scicat_client, derived_dataset):
    dset = deepcopy(derived_dataset)
    dset.pid = PID(prefix=None, pid=dset.pid.pid)
    dset.owner = "a new owner to trigger a failure"
    with pytest.raises(ScicatCommError):
        scicat_client.create_dataset_model(dset)


def test_create_first_orig_datablock(scicat_client, derived_dataset):
    dset = deepcopy(derived_dataset)
    dset.pid = PID(pid="new-dataset-id")
    uploaded = scicat_client.create_dataset_model(dset)

    dblock = OrigDatablock(
        id=PID(prefix="PID.SAMPLE.PREFIX", pid="new-datablock-id"),
        size=9235,
        dataFileList=[
            DataFile(
                path="data.nxs", size=9235, time=parse_date("2023-08-18T13:52:33.000Z")
            )
        ],
        datasetId=uploaded.pid,
        ownerGroup="uu",
        accessGroups=["group1", "2nd_group"],
    )
    scicat_client.create_orig_datablock(dblock)
    downloaded = scicat_client.get_orig_datablocks(uploaded.pid)
    assert len(downloaded) == 1
    downloaded = downloaded[0]
    for key, expected in dblock.dict(exclude_none=True).items():
        # The database populates a number of fields that are None in dset.
        # But we don't want to test those here as we don't want to test the database.
        if key != "dataFileList":
            assert expected == downloaded.dict()[key], f"key = {key}"
    for i in range(len(dblock.dataFileList)):
        for key, expected in dblock.dataFileList[i].dict(exclude_none=True).items():
            assert (
                expected == downloaded.dataFileList[i].dict()[key]
            ), f"i = {i}, key = {key}"


def test_get_dataset(client, derived_dataset, orig_datablock):
    dset = client.get_dataset(derived_dataset.pid)

    assert dset.source_folder == derived_dataset.sourceFolder
    assert dset.creation_time == derived_dataset.creationTime
    assert dset.access_groups == derived_dataset.accessGroups
    assert dset.meta["temperature"] == derived_dataset.scientificMetadata["temperature"]
    assert dset.meta["data_type"] == derived_dataset.scientificMetadata["data_type"]

    for dset_file, expected_file in zip(dset.files, orig_datablock.dataFileList):
        assert dset_file.local_path is None
        assert dset_file.size == expected_file.size
        assert dset_file.creation_time == expected_file.time


def test_fake_can_disable_functions():
    client = FakeClient(
        disable={
            "get_dataset_model": RuntimeError("custom failure"),
            "get_orig_datablocks": IndexError("custom index error"),
        }
    )
    with pytest.raises(RuntimeError, match="custom failure"):
        client.scicat.get_dataset_model(PID(pid="some-pid"))
    with pytest.raises(IndexError, match="custom index error"):
        client.scicat.get_orig_datablocks(PID(pid="some-pid"))


def test_can_get_public_dataset_without_login(request, scicat_access, scicat_backend):
    skip_if_not_backend(request)
    client = Client.without_login(url=scicat_access.url)
    dset = client.get_dataset("PID.SAMPLE.PREFIX/public-dataset")
    assert dset.type == DatasetType.RAW
    assert dset.owner == "librarian"
    assert dset.creation_time == parse_date("2022-11-25T22:12:04.512Z")
    assert dset.number_of_files == 0


def test_cannot_upload_without_login(
    request, derived_dataset, scicat_access, scicat_backend
):
    skip_if_not_backend(request)
    client = Client.without_login(url=scicat_access.url)
    derived_dataset.pid = None
    with pytest.raises(ScicatCommError):
        client.scicat.create_dataset_model(derived_dataset)


def test_get_broken_dataset(request, scicat_access, scicat_backend):
    skip_if_not_backend(request)
    client = Client.without_login(url=scicat_access.url)
    dset = client.get_dataset("PID.SAMPLE.PREFIX/partially-broken")
    assert dset.type == DatasetType.RAW
    # Intact field; was not converted to RemotePath
    assert isinstance(dset.source_folder, str)
    assert dset.source_folder == "/remote/source"

    # Broken fields loaded
    assert isinstance(dset.end_time, str)
    assert dset.end_time == "yesterday"


def test_get_broken_dataset_strict_validation(request, scicat_access, scicat_backend):
    skip_if_not_backend(request)
    client = Client.without_login(url=scicat_access.url)
    with pytest.raises(pydantic.ValidationError):
        client.get_dataset("PID.SAMPLE.PREFIX/partially-broken", strict_validation=True)


def test_cannot_pickle_client_credentials_manual_token_str():
    client = Client.from_token(url="/", token="the-token")  # noqa: S106
    with pytest.raises(TypeError):
        pickle.dumps(client)


def test_cannot_pickle_client_credentials_manual_token_secret_str():
    client = Client.from_token(url="/", token=SecretStr("the-token"))
    with pytest.raises(TypeError):
        pickle.dumps(client)


def test_cannot_pickle_client_credentials_login(request, scicat_access, scicat_backend):
    skip_if_not_backend(request)
    client = Client.from_credentials(
        url=scicat_access.url, **scicat_access.functional_credentials
    )
    with pytest.raises(TypeError):
        pickle.dumps(client)


def test_connection_error_does_not_contain_token():
    client = Client.from_token(
        url="https://not-actually-a_server",
        token="the token/which_must-be.kept secret",  # noqa: S106
    )
    try:
        client.get_dataset("does not exist")
        assert False, "There must be an exception"  # noqa: B011
    except Exception as exc:
        assert "the token/which_must-be.kept secret" not in str(exc)
        for arg in exc.args:
            assert "the token/which_must-be.kept secret" not in str(arg)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)

"""Validator for ORCIDs.

Based on
https://support.orcid.org/hc/en-us/articles/360006897674-Structure-of-the-ORCID-Identifier
"""


def orcid_checksum(orcid_id: str) -> str:
    total = 0
    for c in orcid_id.replace("-", "")[:-1]:
        total = (total + int(c)) * 2
    result = (12 - total % 11) % 11
    return "X" if result == 10 else str(result)


def is_valid_orcid(orcid_uri: str) -> bool:
    base, orcid_id = orcid_uri.rsplit("/", 1)
    if base != "https://orcid.org":
        # must start with the ORCID URL
        return False
    if orcid_id.count("-") != 3:
        # must have four blocks of numbers
        return False
    if not all(len(x) == 4 for x in orcid_id.split("-")):
        # each block must have 4 digits
        return False
    if orcid_checksum(orcid_id) != orcid_id[-1]:
        # checksum must match last digit
        return False
    return True



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)
from contextlib import contextmanager
from pathlib import Path
from typing import Any, Dict, Iterator, List, Optional, Union

try:
    from pyfakefs.fake_filesystem import FakeFilesystem
except ImportError:
    FakeFilesystem = Any

from ..dataset import Dataset
from ..file import File
from ..filesystem import RemotePath
from ..transfer.util import source_folder_for


class FakeDownloadConnection:
    def __init__(self, fs: Optional[FakeFilesystem], files: Dict[RemotePath, bytes]):
        self.files = files
        self.fs = fs

    def download_file(self, *, remote: RemotePath, local: Path) -> None:
        if self.fs is not None:
            self.fs.create_file(local, contents=self.files[remote])
        else:
            with open(local, "wb") as f:
                f.write(self.files[remote])

    def download_files(self, *, remote: List[RemotePath], local: List[Path]) -> None:
        for r, l in zip(remote, local):
            self.download_file(remote=r, local=l)


class FakeUploadConnection:
    def __init__(
        self,
        files: Dict[RemotePath, bytes],
        reverted: Dict[RemotePath, bytes],
        source_folder: RemotePath,
    ):
        self.files = files
        self.reverted = reverted
        self._source_folder = source_folder

    def _remote_path(self, filename: RemotePath) -> RemotePath:
        return self._source_folder / filename

    def _upload_file(self, *, remote: RemotePath, local: Path) -> RemotePath:
        remote = self._remote_path(remote)
        with open(local, "rb") as f:
            self.files[remote] = f.read()
        return remote

    def upload_files(self, *files: File) -> List[File]:
        for file in files:
            self._upload_file(
                remote=file.remote_path, local=file.local_path
            )  # type: ignore[arg-type]
        return list(files)

    def revert_upload(self, *files: File) -> None:
        for file in files:
            remote = self._remote_path(file.remote_path)
            self.reverted[remote] = self.files.pop(remote)


class FakeFileTransfer:
    """Mimic a file down-/upload handler.

    Files are not transferred to a server but loaded into an internal storage.

    It is possible to use a fake file system as implemented by
    `pyfakefs <https://pypi.org/project/pyfakefs/>`_.

    Examples
    --------
    Using the ``fs`` fixture from pyfakefs in pytest:

    .. code-block:: python

        def test_upload(fs):
            client = FakeClient.from_token(
                url="...",
                token="...",
                file_transfer=FakeFileTransfer(fs=fs))
            dset = ...
            client.upload_new_dataset_now(dset)
            assert client.file_transfer.files[expected_remote_path] == file_content

    See Also
    --------
    scitacean.testing.client.FakeClient:
        Client to mimic a SciCat server.
    """

    def __init__(
        self,
        *,
        fs: Optional[FakeFilesystem] = None,
        files: Optional[Dict[Union[str, RemotePath], bytes]] = None,
        reverted: Optional[Dict[Union[str, RemotePath], bytes]] = None,
        source_folder: Optional[Union[str, RemotePath]] = None,
    ):
        """Initialize a file transfer.

        Parameters
        ----------
        fs:
            Fake filesystem. If given, files are down-/uploaded to/from this
            filesystem instead of the real one.
            If set to ``None``, the real filesystem is used.
        files:
            Initial files stored "on the server".
            Maps file names to contents.
        reverted:
            Files that have been uploaded and subsequently been removed.
        source_folder:
            Upload files to this folder if set.
            Otherwise, upload to the dataset's source_folder.
            Ignored when downloading files.
        """
        self.fs = fs
        self.files = _remote_path_dict(files)
        self.reverted = _remote_path_dict(reverted)
        self._source_folder_pattern = source_folder

    def source_folder_for(self, dataset: Dataset) -> RemotePath:
        return source_folder_for(dataset, self._source_folder_pattern)

    @contextmanager
    def connect_for_download(self) -> Iterator[FakeDownloadConnection]:
        yield FakeDownloadConnection(fs=self.fs, files=self.files)

    @contextmanager
    def connect_for_upload(self, dataset: Dataset) -> Iterator[FakeUploadConnection]:
        yield FakeUploadConnection(
            files=self.files,
            reverted=self.reverted,
            source_folder=self.source_folder_for(dataset),
        )


def _remote_path_dict(
    d: Optional[Dict[Union[str, RemotePath], bytes]]
) -> Dict[RemotePath, bytes]:
    if d is None:
        return {}
    return {RemotePath(path): contents for path, contents in d.items()}



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)

from dataclasses import dataclass
from pathlib import Path

import pytest
import yaml

from .docker import docker_compose

_SSH_SERVER_DOCKER_CONFIG = (
    Path(__file__).resolve().parent / "docker-compose-ssh-server.yaml"
)


@dataclass
class SSHAccess:
    host: str
    port: int
    username: str
    password: str


def _load_config() -> SSHAccess:
    with open(_SSH_SERVER_DOCKER_CONFIG, "r") as f:
        config = yaml.safe_load(f)
    service = config["services"]["scitacean-test-ssh-server"]
    env = {k: v for k, v in map(lambda s: s.split("="), service["environment"])}
    return SSHAccess(
        host="localhost",
        port=service["ports"][0].split(":")[0],
        username=env["USER_NAME"],
        password=env["USER_PASSWORD"],
    )


@pytest.fixture(scope="session")
def ssh_access():
    return _load_config()


@pytest.fixture(scope="session")
def ssh_fileserver(request, ssh_access):
    """Spin up an SSH server.

    Does nothing unless the --ssh-tests command line option is set.
    """
    if not request.config.getoption("--ssh-tests"):
        yield False

    with docker_compose(_SSH_SERVER_DOCKER_CONFIG):
        yield True
